//
//  TunerScene.cpp
//  GlobalGameJam17
//
//  Created by Cesar Parent on 21/01/2017.
//  Copyright (c) 2017 teamrocket. All rights reserved.
//
#include "TunerScene.hpp"
#include "GameObject.hpp"
#include "ScoreMenu.hpp"
#include "CollisionManager.hpp"
#include "ScoreManager.hpp"
#include "InfRunner.hpp"
#include "Bullet.hpp"
#include "Enemy.hpp"
#include "Tick.hpp"
#include "Utils.hpp"

enum Layers {
    BACKGROUND,
    TICKS,
    PLAYER,
    BULLET,
    MONSTER,
    UI,
    NEEDLE,
    OVERLAY,
};

enum Category {
    CAT_PLAYER  = 1 << 0,
    CAT_TICK    = 1 << 1,
    CAT_ENEMY   = 1 << 2,
    CAT_BULLET  = 1 << 3,
};

TunerScene::TunerScene()
: score_(0)
, flexSeconds_(0)
, nextEnemy_(false)
, nextTick_(true) {
    setAnchorPoint(0.f, 0.f);
    setZSort(true);
    
    Constants::reload();
    InfRunner::instance().start();
    
    setBackgroundColor(Color{.1f, .1f, .1f});
    auto back = alloc<SpriteNode>("assets/sprites/background.png");
    back->setAnchorPoint(0.f, 0.f);
    back->setZPosition(Layers::BACKGROUND);
    addChild(back);
    
    front_ = alloc<SpriteNode>("assets/sprites/foreground.png");
    front_->setAnchorPoint(0.f, 0.f);
    front_->setZPosition(Layers::OVERLAY);
    addChild(front_);
    
    bounds_ = Rect(0.f, 135.f, size().width, 30.f).inset(76.f, 0.f);
    
    player_ = alloc<Player>(bounds_.maxY());
    player_->setPosition(bounds_.minX() + 30.f, bounds_.maxY());
    player_->setZPosition(Layers::PLAYER);
    player_->categoryMask = CAT_PLAYER;
    player_->collisionMask = CAT_TICK | CAT_ENEMY;
    CollisionManager::instance().addObject(player_);
    addChild(player_);
    
    needle_ = alloc<SpriteNode>("assets/sprites/needle.png");
    needle_->setAnchorPoint(.5f, 0.f);
    needle_->setPosition(bounds_.minX(), bounds_.minY());
    needle_->setZPosition(Layers::NEEDLE);
    
    addChild(needle_);
    
    NotificationCenter::defaultCenter().addObserverWithName(this,
                                                            selO(TunerScene::playerDied),
                                                            Notif::DEATH,
                                                            nullptr);
    NotificationCenter::defaultCenter().addObserverWithName(this,
                                                            selO(TunerScene::explosion),
                                                            Notif::EXPLOSION,
                                                            nullptr);
    
    pauseLabel_ = PlistDecoder::decodeObjectFromFile<LabelNode>("assets/objects/label.plist");
    pauseLabel_->setText("paused");
    pauseLabel_->setFontColor(Constants::OVERLAY_COLOUR);
    pauseLabel_->setPosition(bounds_.center().x, bounds_.maxY() + 20.f);
    pauseLabel_->setHidden(true);
    pauseLabel_->setZPosition(Layers::UI);
    addChild(pauseLabel_);
    
    
    scoreLabel_ = PlistDecoder::decodeObjectFromFile<LabelNode>("assets/objects/label.plist");
    scoreLabel_->setText("000000");
    scoreLabel_->setFontColor(Constants::OVERLAY_COLOUR);
    scoreLabel_->setHorizontalAlignmentMode(LabelNode::HorizontalAlignmentMode::Left);
    scoreLabel_->setPosition(bounds_.minX() + 30.f, bounds_.maxY() + 20.f);
    scoreLabel_->setZPosition(Layers::UI);
    addChild(scoreLabel_);
    
}

TunerScene::~TunerScene() {
    NotificationCenter::defaultCenter().removeObserver(this);
}

void TunerScene::didMoveToView() {
    Controls::setDelegate(this);
    runAction(Action::sequence({
        Action::waitForDuration(5.f),
        Action::runBlock([this]{
            nextEnemy_ = true;
        })
    }), "waitabit");
}


void TunerScene::update(double delta) {
    if(InfRunner::instance().gameOver()) return;
    
    CollisionManager::instance().checkCollisions();
    addTick();
    addEnemy();
    addBullet();
    manageScore(delta);
    needle_->setX(lerp(bounds_.minX(), bounds_.maxX(), InfRunner::instance().updateNeedle()));
}

void TunerScene::togglePause() {
    if(this->paused()) {
        setPaused(false);
        pauseLabel_->setHidden(true);
    } else {
        pauseLabel_->setHidden(false);
        setPaused(true);
    }
}

// MARK: - ControlDelegate

void TunerScene::keyDown(Meteor::Key aKey) {
    switch (aKey) {
        case Key::SPACE:
            arc_cast<Player>(player_)->controls.shoot = true;
            break;
        case Key::UP:
            arc_cast<Player>(player_)->controls.jump = true;
            break;
        case Key::LEFT:
            InfRunner::instance().controls.left = true;
            break;
        case Key::RIGHT:
            InfRunner::instance().controls.right = true;
            break;
        case Key::BACKSPACE:
            Constants::reload();
            break;
        case Key::ESCAPE:
            togglePause();
            break;
        default:
            break;
    }
}

void TunerScene::keyUp(Meteor::Key aKey) {
    switch (aKey) {
        case Key::SPACE:
            arc_cast<Player>(player_)->controls.shoot = false;
            break;
        case Key::UP:
            arc_cast<Player>(player_)->controls.jump = false;
            break;
        case Key::LEFT:
            InfRunner::instance().controls.left = false;
            break;
        case Key::RIGHT:
            InfRunner::instance().controls.right = false;
            break;
        default:
            break;
    }
}

void TunerScene::controllerButtonDown(int controller, Button aButton) {
    switch (aButton) {
        case Button::A:
            player_->controls.jump = true;
            break;
        case Button::RIGHTSHOULDER:
        case Button::LEFTSHOULDER:
            player_->controls.shoot = true;
            break;
        case Button::DPAD_LEFT:
            InfRunner::instance().controls.left = true;
            break;
        case Button::DPAD_RIGHT:
            InfRunner::instance().controls.right = true;
            break;
        case Button::START:
            togglePause();
            break;
        default:
            break;
    }
}

void TunerScene::controllerButtonUp(int controller, Button aButton) {
    switch (aButton) {
        case Button::A:
            player_->controls.jump = false;
            break;
        case Button::RIGHTSHOULDER:
        case Button::LEFTSHOULDER:
            player_->controls.shoot = false;
            break;
        case Button::DPAD_LEFT:
            InfRunner::instance().controls.left = false;
            break;
        case Button::DPAD_RIGHT:
            InfRunner::instance().controls.right = false;
        default:
            break;
    }
}

void TunerScene::controllerAxisMoved(int controller, Axis theAxis, float value) {
    if(theAxis != Axis::LEFTX) return;
    InfRunner::instance().controls.left = value < -0.5f;
    InfRunner::instance().controls.right = value > .5f;
}

// MARK: - Notification Receivers

void TunerScene::playerDied(Notification* notification) {
    if(InfRunner::instance().gameOver()) return;
    
    // Add the remaining score.
    score_ += flexSeconds_ * Constants::SCORE_PER_SECOND;
    ScoreManager::instance().addScore(score_);
    CollisionManager::instance().removeAll();
    
    InfRunner::instance().loose();
    runAction(Action::sequence({
        Action::waitForDuration(1.f),
        Action::runBlock([this]{
            Game::instance().presentScene(alloc<ScoreMenu>(score_),
                                          Transition::fadeWithDuration(1.f));
        })
    }), "gameOver");
}

void TunerScene::explosion(Notification* notification) {
    if(InfRunner::instance().gameOver()) return;
    AudioEngine::playSound("assets/audio/explosion.wav");
    
    front_->runAction(Action::customActionWithBlock([](Node* node, float t){
        vec2 offset{Utils::random(-3.f, 3.f), Utils::random(-3.f, 3.f)};
        node->setPosition((1.f-t) * offset);
    }, 1.f), "shake");
}

// MARK: - Level Management

void TunerScene::addTick() {
    if(!nextTick_) return;
    nextTick_ = false;
    
    auto obj = alloc<Tick>(3.f * Utils::random<int>(2, 4) + 18.f);
    obj->setName(Constants::TICK_TAG);
    obj->setPosition(bounds_.maxX(), bounds_.maxY() + 18.f);
    obj->setZPosition(Layers::TICKS);
    obj->setColor(Constants::OVERLAY_COLOUR);
    
    CollisionManager::instance().addObject(obj);
    addChild(obj);
    
    obj->categoryMask = CAT_TICK;
    obj->collisionMask = 0;
    
    runAction(Action::sequence({
        Action::waitForDuration(Utils::random(1.f, 2.f) / InfRunner::instance().speedFactor()),
        Action::runBlock([this]{
            nextTick_ = true;
        })
    }), "tick");
}

void TunerScene::addEnemy() {
    if(!nextEnemy_) return;
    nextEnemy_ = false;
    
    auto obj = alloc<Enemy>("assets/sprites/placeholder.png", bounds_.center().y);
    obj->setName(Constants::ENEMY_TAG);
    obj->setPosition(bounds_.maxX(), bounds_.center().y);
    obj->setZPosition(Layers::MONSTER);
    obj->setColor(Constants::OVERLAY_COLOUR);
    //obj->setScript("assets/scripts/enemy.code");
    
    CollisionManager::instance().addObject(obj);
    addChild(obj);
    
    obj->categoryMask = CAT_ENEMY;
    obj->collisionMask = CAT_BULLET;
    
    runAction(Action::sequence({
        Action::waitForDuration(Utils::random(1.f, 3.f) / InfRunner::instance().speedFactor()),
        Action::
        Action::runBlock([this]{
            nextEnemy_ = true;
        })
    }), "spawn");
}

void TunerScene::addBullet() {
    if(!player_->canShoot()) return;
    
    AudioEngine::playSound("assets/audio/shoot.wav");
    
    auto obj = alloc<Bullet>(bounds_.maxX());
    obj->setPosition(player_->frame().center());
    
    obj->categoryMask = CAT_BULLET;
    obj->collisionMask = CAT_TICK;
    
    CollisionManager::instance().addObject(obj);
    addChild(obj);
}

void TunerScene::manageScore(double delta) {
    flexSeconds_ += delta * std::pow(InfRunner::instance().speedFactor(), 2.f);
    if(flexSeconds_ < 1.f) return;
    score_ += Constants::SCORE_PER_SECOND;
    flexSeconds_ -= 1.f;
    
    static char scoreString[8];
    std::sprintf(scoreString, "%06d", score_);
    scoreLabel_->setText(scoreString);
}
