//
//  NotificationCenter.h
//  Meteor
//
//  Created by Cesar Parent on 13/08/2015.
//  Copyright (c) 2015 Cesar Parent. All rights reserved.
//

#ifndef __Meteor__NotificationCenter__
#define __Meteor__NotificationCenter__

#include "../defines.h"
#include "../CPFoundation.h"
#include <vector>
#include <string>

NS_BEGIN

/*
 * @defgroup        Notification Notification Types
 * @brief           List of existing notification types
 */

struct Notification : CPObject {
    
    Notification(const char* name, CPObject* sender) : name(name), sender(sender){}
    
    const char* name;
    CPObject* sender;
};


/*!
 * @class           NotificationCenter
 * @ingroup         Foundation Management
 * @brief           A class that provides a mechanism for broadcasting information within a program.
 * @details         Functions can be registered with the Notification Center to receive
 *                  notifications (Notification objects) using addObserver(). A method can be
 *                  registered multiple times, and can thus receive multiple types of notifications.
 *
 *                  While you can create instances of NotificationCenter, most uses cases are
 *                  covered by the default singletong instance.
 */
class NotificationCenter final {
public:
    
    template<typename T>
    struct Payload : public CPObject {
        Payload(T v1 = T(), T v2 = T()) : data1(v1), data2(v2) {}
        
        T data1;
        T data2;
    };
    
    typedef Payload<int> IntPayload;
    typedef Payload<float> FloatPayload;
    
    static NotificationCenter& defaultCenter();
    
    NotificationCenter();
    
    ~NotificationCenter();
    
    /**
     * @name        Registering Observers
     * @{
     */
    
    /*!
     * @brief       Registers an observer function for a notification type and sender.
     *              If <code>name</code> is <code>nullptr</code>, then all notifications with a
     matching <code>object</code> will be forwarded to <code>observer</code>. If
     *              <code>object</code> is <code>null</code>, all notifications with a matching name
     *              will be. If both conditions are true, then every notification will be forwarded.
     * @param       target      The object that will receive notifications.
     * @param       selector    CPObject* selector called when a notification is posted.
     * @param       name        Name of the notifications the observer wants to receive.
     * @param       sender      The object to receive notifications from, or nullptr.
     */
    void addObserverWithName(CPObject* target,
                             selector_o selector,
                             const char* name,
                             CPObject* sender);
    
    /**
     * @}
     * @name        Removing Observers
     * @{
     */
    
    /*!
     * @brief       Removes observers for the matching name and object. Any object that will be
     *              deleted should remove itself from the notification center's dispatch list.
     * @param       target      Observer to un-register.
     * @param       name        Notification name of the observer to remove.
     */
    void removeObserver(CPObject* target, const char* notification = nullptr);
    
    
    /**
     * @}
     * @name        Posting Notifications
     * @{
     */
    
    /*!
     * @brief       Posts a notification to matchin observers.
     * @param       name        Notification's name.
     * @param       sender      The object to attach to the notification. This is usually the object
     *                          sending the notification.
     */
    void postNotificationWithName(const char* name, CPObject* sender);
    
    /**
     * @}
     */
    
private:
    
    bool _observerExists(CPObject* obs, const char* name, CPObject* sender);
    
    struct _observer {
        
        CPObject*   target;
        selector_o  selector;
        const char* name;
        CPObject*   sender;
        
        void performSelector(CPObject* obj = nullptr) const {
            (target->*selector)(obj);
        }
    };
    
    std::vector<_observer> _observers;
    
    NotificationCenter(const NotificationCenter&) = delete;
    NotificationCenter(const NotificationCenter&&) = delete;
    void operator=(const NotificationCenter&) = delete;
};

NS_END

#endif /* defined(__Meteor__NotificationCenter__) */
